#!/usr/bin/perl -w

##### DESCRIPTION ######################################################
# genfs2.pl Will generate a complex/deep directory structure consisting
# random directories and files.
#
# (C) 2015 Preston de Guise
# You may modify this script as you wish, but you may not charge for it,
# and you must not remove this comment block.
########################################################################

##### MODULES ##########################################################
use strict;
use Getopt::Std;
use File::Basename;
########################################################################

##### LOREM IPSUM ######################################################
my @lorem=qw/lorem ipsum dolor sit amet consectetur adipiscing elit aenean rhoncus urna at sapien laoreet ac bibendum semper proin vestibulum ligula in est scelerisque lobortis id cras dictum sem aliquam erat volutpat nulla facilisi donec aliquet nisl euismod lacinia leo vitae venenatis mi quis nullam odio faucibus hendrerit eu posuere a tellus libero tortor tincidunt commodo ut magna orci condimentum in eleifend facilisis sed mauris felis ornare eget ullamcorper ante duis nisi at metus rutrum et justo vehicula velit vel dignissim tristique viverra purus nec risus fringilla interdum et a nam lectus praesent massa fermentum in tempus iaculis augue auctor neque porta suspendisse enim dapibus pulvinar pellentesque habitant morbi senectus netus malesuada fames turpis egestas ultricies varius non nibh/;
########################################################################

my $minDirsPerLayer = 5;
my $maxDirsPerLayer = 10;
my $minFilesPerLayer = 5;
my $maxFilesPerLayer = 10;
my $minRecurse = 5;
my $maxRecurse = 10;
my $targetDir = "not supplied";
my $minFileSize = 1024;
my $maxFileSize = 1048576;
my $dataChunkSize = 52428800;
my $minFileNameLength = 5;
my $maxFileNameLength = 15;
my $compressibleFiles = 0;
my $minimumPermittedFileSize = 511;
my $chunkSize = 8192;
my $segSize = 512;
my $tmp;
my $loremIpsumNames = 0;
my @chars=("a","b","c","d","e","f","g","h","i","j","k","l","m","n","o",
		   "p","q","r","s","t","u","v","w","x","y","z","_","@","=","+",
		   "A","B","C","D","E","F","G","H","I","J","K","L","M","N","O",
		   "P","Q","R","S","T","U","V","W","X","Y","Z","1","2","3","4",
		   "5","6","7","8","9");
my $charLength = @chars+0;
my $randomChars;
my @extensions = (".ppt",".doc",".docx",".txt",".dat",".idx",".anr",".bsb",
	 ".xls",".xlm",".xlt",".docm",".zip",".tar",".tgz",".tbz2",".tar.bz2",
	 ".tar.gz",".rar",".mp3",".mov",".wma",".wmv",".dot",".zap",".7r",".hqx",
	 ".spa",".gba");
my $extCount = @extensions+0;
my $currentYear = (localtime(time))[5]+1900 - 2000;
my $maxFileCount = -1;
my $currentFileCount = 0;
my $quiet = 0;
my $self = basename($0);
my %opts;


##############################################################################
# Subroutines
##############################################################################

# in_mb($bytes) returns the size of a given number in MB.
sub in_mb {
	if (@_+0 != 1) {
		die "Unexpected input for in_mb(bytes)\n";
	}
	
	# else...
	my $bytes = $_[0];
	if (defined($bytes)) {
		return $bytes / 1024 / 1024;
	} else {
		die "in_mb: Got null bytes.\n";
	}
}

# get_options() retrieves the options and borks if necessary.
sub get_options {
	if (getopts('hd:D:f:F:r:R:t:s:S:l:L:C:P:T:qI',\%opts)) {
		usage() if (defined($opts{h}));
		if (defined($opts{d})) {
			if ($opts{d} < 1) {
				die "$opts{d} is too small for minimum number of directories.\n";
			} else {
				$minDirsPerLayer = $opts{d};
			}
		}
		if (defined($opts{T})) {
			if ($opts{T} =~ /^\d+$/) {
				$maxFileCount = $opts{T};
			} else {
				die "Maximum total files (-T) must be expressed as a positive integer.\n";
			}
		}
		if (defined($opts{I})) {
			$loremIpsumNames = 1;
		}
		if (defined($opts{q})) {
			$quiet = 1;
		}
		if (defined($opts{D})) {
			if ($opts{D} < $minDirsPerLayer) {
				die "Max dirs per layer must be larger than min dirs per layer.\n";
			} else {
				$maxDirsPerLayer = $opts{D};
			}
		} 
		if (defined($opts{f})) {
			if ($opts{f} < 1) {
				die "$opts{f} is too small for minimum number of files.\n";
			} else {
				$minFilesPerLayer = $opts{f};
			}
		}
		if (defined($opts{F})) {
			if ($opts{F} < $minFilesPerLayer) {
				die "Max files per layer must be larger than min files per layer.\n";
			} else {
				$maxFilesPerLayer = $opts{F};
			}
		} 
		if (defined($opts{r})) {
			if ($opts{r} < 1) {
				die "Minimum recursion depth is 1.\n";
			} else {
				$minRecurse = 1;
			}
		}
		if (defined($opts{R})) {
			if ($opts{R} < $minRecurse) {
				die "Maximum recursion depth must be deeper than minimum recursion depth.\n";
			} else {
				$maxRecurse = $opts{R};
			}
		}
		if (defined($opts{t})) {
			if (-d $opts{t}) {
				$targetDir = $opts{t};
			} else {
				die "Specified target directory $opts{t} does not exist!\n";
			}
		} else {
			die "Target directory must be specified (and exist)\n";
		}
		if (defined($opts{s})) {
			if ($opts{s} < $minimumPermittedFileSize) {
				die "Minimum file size must be at least $minimumPermittedFileSize bytes.\n";
			} else {
				$minFileSize = $opts{s};
			}
		}
		if (defined($opts{S})) {
			if ($opts{S} < $minFileSize) {
				die "Maximum filesize must be larger than minimum filesize.\n";
			} else {
				$maxFileSize = $opts{S};
			}
		}
		if (defined($opts{l})) {
			if ($opts{l} < 2) {
				die "The minimum filename length must be >= 2.\n";
			} else {
				if (defined($opts{N})) {
					warn("Semi-realistic filenames (-N) requested. Min filename length will be ignored.");
				}
				$minFileNameLength = $opts{l};
			}
		}
		if (defined($opts{L})) {
			if ($opts{L} < $minFileNameLength) {
				die "The maximum filename length must be greater than the minimum filename length.\n";
			} else {
				if (defined($opts{N})) {
					warn("Semi-realistic filenames (-N) requested. Max filename length will be ignored.");
				}
				$maxFileNameLength = $opts{L};
			}
		}
		if (defined($opts{C})) {
			$compressibleFiles = 1;
		}
		if (defined($opts{P})) {
			$dataChunkSize = $opts{P};
		}
	} else {
		die "Many options were not present (fix this error message)\n";
	}
}

# randfile($size,$filename) generates a random file from $randomChars using
# $chunkSize and $segSize to more quickly generate it.
sub randfile {
	my ($size,$filename) = @_;
	if (-f $filename) {
		return 0;
	}
	if ($size < 1) {
		return 0;
	}
	my $currSize = 0;

	# else...
	my $segsPerChunk = int($chunkSize / $segSize);
	my $i;
	my $dataSize = length($randomChars);
	my $start;

	if (open(FILE,">$filename")) {
		while ($currSize < $size) {
			my $chunk = "";
			for ($i = 0; $i < $segsPerChunk; $i++) {
				$start = int(rand($dataSize));
				$chunk .= substr($randomChars,$start,$segSize);
			}
			if (length($chunk) > ($size-$currSize)) {
				$chunk = substr($chunk,0,$size-$currSize);
			}
			syswrite(FILE,$chunk);
			$currSize += length($chunk);
		}
		close(FILE);
		return 1;
	} else {
		return 0;
	}
}


# generate_files($dir) generates a random number of files within the given
# bounds ($minFilesPerLayer -> $maxFilesPerLayer)
# Returns the number of files generated, or 0 if there's an error.
sub generate_files {
	my $dir = $_[0];
	return 0 if (! -d $dir);

	my $numFiles = int(rand($maxFilesPerLayer - $minFilesPerLayer)) + $minFilesPerLayer;
	(!$quiet) && print "($numFiles).\n";
	my $count;
	for ($count=0; $count<$numFiles; $count++) {
		my $filename = rand_filename($minFileNameLength,$maxFileNameLength);
		# Now, since this is a filename, quickly work out whether we should attach an
		# extension to the file.
		my $extensionRequired = int(rand(6));
		if ($extensionRequired != 0) {
			my $extension = $extensions[(int(rand(@extensions+0)))];
			$filename .= $extension;
		}
		my $fileSize = int(rand($maxFileSize-$minFileSize))+$minFileSize;
		#print "\tWill generate $filename ($fileSize bytes) in $dir\n";
		randfile($fileSize,"$dir/$filename");
		$currentFileCount++;
		if ($currentFileCount % 100 == 0) {
			print "\t --- $currentFileCount files\n";
		}
		if ($maxFileCount > 0 && $currentFileCount >= $maxFileCount) {
			# finish up here.
			print "\n\nHit maximum file count ($maxFileCount).\n";
			die "Terminating.\n";
		}
	}
}

# generate_subdirs($min,$max) generates a list of subdirectory names -
# with a minimum of $min and a maximum of $max.
sub generate_subdirs {
	my ($min, $max) = @_;
	my $numDirs = int(rand($max-$min))+$min;
	my $count;
	my @subDirList;
	for ($count=0; $count<$numDirs; $count++) {
		push(@subDirList,rand_filename($minFileNameLength,$maxFileNameLength));
	}
	return @subDirList;
}

# populate_subdir($dirName,$recurseDepth) populates a given $dirName (creating
# it) with files and subdirectories to a maximum depth of $recurseDepth.
sub populate_subdir {
	my ($dirName,$recurseDepth) = @_;
	# TODO: Add error checking here.
	mkdir($dirName,0775);
	(!$quiet) && print "\tGenerating files for $dirName. ";
	generate_files($dirName);
	return 1 if ($recurseDepth == 0);

	# else...
	my @subDirs=generate_subdirs($minDirsPerLayer,$maxDirsPerLayer);
	my $subDir;
	foreach $subDir (@subDirs) {
		my $subRecurseDepth = int(rand($recurseDepth-1));
		(!$quiet) && print "\tWill populate $dirName/$subDir to max depth $subRecurseDepth\n";
		populate_subdir("$dirName/$subDir",$subRecurseDepth);
	}
}

# rand_filename($min,$max) generates a random filename at least $min
# chars long and at most $max chars long. If the "loremIpsumNames" flag
# is turned on the min/max details are ignored.
sub rand_filename {
	my ($min,$max) = @_;
	my $charListLength = @chars+0;
	my $length = int(rand($max-$min))+$min;
	my $filename = "";
	if ($loremIpsumNames) {
		my $nameCount = int(rand(7))+1;
		foreach (my $count = 1; $count <= $nameCount; $count++) {
			$filename .= $lorem[int(rand(@lorem+0))-1];
			if ($count != $nameCount) {
				if (int(rand(4)) > 1) {
					$filename .= "-";
				} else {
					$filename .= " ";
				}
			}
		}
	} else {
		for (my $count = 0; $count<$length; $count++) {
			$filename .= $chars[int(rand($charListLength))];
		}
	}
	return $filename;
}

# pre_generate($size) pre-generates an array containing random chars
# to a maximum of $size bytes. The array is assumed to be @randomChars
# because we don't want to be passing say, a 500MB an array backwards
# and forwards
sub pre_generate {
	my $size = $_[0];
	my $count;
	my $max = ($compressibleFiles == 0) ? 255 : 97;
	my $number;
	my $lastPercent = -1;
	$size += ($segSize*2);	# Give ourselves a bit of buffer.
	for ($count = 1; $count <= $size; $count++) {
		$randomChars .= chr(int(rand($max)));
		$number = int($count / $size * 100);
		if (!$quiet) {
			if ($number % 10 == 0) {
				print "\t\t$number% of random data chunk generated.\n" if ($lastPercent != $number);
				$lastPercent = $number;
			}
		}
	}
}


# usage() gives usage information and exits.
sub usage {
	die "Syntax: $self [-d minDir] [-D maxDir] [-f minFile] [-F maxFile] [-r minRecurse] [-R maxRecurse] -t target [-s minSize] [-S maxSize] [-l minLength] [-L maxLength] [-C] [-P dCsize] [-T mfc] [-q] [-I]\n

Creates a randomly populated directory structure for backup/recovery 
and general performance testing. Files created are typically non-
compressible.

All options other than target are optional. Values in parantheses beside
explanations denote defaults that are used if not supplied.

Where:

	-d minDir\tMinimum number of directories per layer. (5)
	-D maxDir\tMaximum number of directories per layer. (10)
	-f minFile\tMinimum number of files per layer. (5)
	-F maxFile\tMaximum number of files per layer. (10)
	-r minRecurse\tMinimum recursion depth for base directories. (5)
	-R maxRecurse\tMaximum recursion depth for base directories. (10)
	-t target\tTarget where directories are to start being created.
	\t\tTarget must already exist. This option MUST be supplied.
	-s minSize\tMinimum file size (in bytes). (1 K)
	-S maxSize\tMaximum file size (in bytes). (1 MB)
	-l minLength\tMinimum filename/dirname length. (5)
	-L maxLength\tMaximum filename/dirname length. (15)
	-P dCsize\tPre-generate random data-chunk at least dcSize bytes.
	\t\tWill default to $dataChunkSize bytes.
	-C \t\tTry to provide compressible files.
	-I \t\tUse lorem ipsum filenames.
	-T mfc\t\tSpecify maximum number of files that will be created.
	\t\tDoes not include directories in count.
	-q\t\tQuiet mode. Only print updates to the file-count.

E.g.:

$0 -r 2 -R 32 -s 512 -S 65536 -t /d/06/test

Would generate a random filesystem starting in /d/06/test, with a minimum
recursion depth of 2 and a maximum recursion depth of 32, with a minimum
filesize of 512 bytes and a maximum filesize of 64K.\n\n";
}


##############################################################################
# Main
##############################################################################

get_options();

# Die if we don't have a target directory specified. Everything else is optional/
# has defaults assigned to it.
die "Target directory (-t) not specified\n" if ($targetDir eq "not supplied");

# Now, generate dump of settings.
if (!$quiet) {
	print "Configured with the following settings:\n";
	print "\tMin dirs per layer = $minDirsPerLayer\n";
	print "\tMax dirs per layer = $maxDirsPerLayer\n";
	print "\tMin files per layer = $minFilesPerLayer\n";
	print "\tMax files per layer = $maxFilesPerLayer\n";
	print "\tTarget directory = $targetDir\n";
	print "\tMin file size = $minFileSize\n";
	print "\tMax file size = $maxFileSize\n";
	print "\tMin recursion depth = $minRecurse\n";
	print "\tMax recursion depth = $maxRecurse\n";
	if (!$loremIpsumNames) {
		print "\tMin filename length = $minFileNameLength\n";
		print "\tMax filename length = $maxFileNameLength\n";
	} else {
		print "\tUse lorem ipsum filenames = yes\n";
	}
	print "\tTry for compressible files = $compressibleFiles\n";
	if ($maxFileCount > -1) {
		print "\tMaximum number of files to be created = $maxFileCount\n";
	}
	print "\tPre-generate a datachunk size = " . in_mb($dataChunkSize) . "MB\n";
	print "\n";
}

print "Progress:\n";
print "\tPre-generating random data chunk. (This may take a while.)\n";
pre_generate($dataChunkSize);

# Start doing the work.
if ($quiet) {
	print "\tGenerating files. Standby.\n";
} else {
	print "\tGenerating files in root directory of $targetDir. ";
}
generate_files($targetDir);
my @subDirList = generate_subdirs($minDirsPerLayer,$maxDirsPerLayer);
my $subDir;
foreach $subDir (@subDirList) {
	my $recurseLength = int(rand($maxRecurse-$minRecurse))+$minRecurse;
	(!$quiet) && print "\tNew subdirectory $subDir - max recurse depth $recurseLength\n";
	populate_subdir("$targetDir/$subDir",$recurseLength);
}

